# The MIF Coordinate Registrator - Version 1.10
#
# Copyright (C) 2001-2008 by Igor Belov
# All Rights Reserved
#
# @(#) Mifreg 1.10 -- 2008-03-20

eval destroy [winfo child .]
wm withdraw .

catch {destroy .calc}
toplevel .calc
wm title .calc "The MIF Coordinate Registrator"
wm iconname .calc "Mifreg"
wm resizable .calc 0 0

set tab_dir "."
set mii_dir "."
set mio_dir "."
set coord 1
set rc "mifreg.ini"
if {[file exists $rc] != 0} then {
    source $rc
}

label .calc.msg -wraplength 4i -justify center -font "Helvetica 10 bold" \
	-text "Register MIF files using raster image registration"

frame .calc.in -bd 2 -relief groove
label .calc.in.tab_lbl -text "Raster registration TAB file:"
entry .calc.in.tab_ent -relief sunken -bg white -width 40
button .calc.in.tab_but -text ". . ." -font "Helvetica 6" -command {
    set s [tk_getOpenFile -filetypes {{{TAB files} {.tab}}} \
	    -initialdir $tab_dir -title "Raster registration TAB file"]
    if {[string length $s] > 0} then {
	.calc.in.tab_ent delete 0 end
	.calc.in.tab_ent insert 0 $s
	set tab_dir [file dirname $s]
    }
}
label .calc.in.mii_lbl -text "Raw MIF file:"
entry .calc.in.mii_ent -relief sunken -bg white -width 40
button .calc.in.mii_but -text ". . ." -font "Helvetica 6" -command {
    set s [tk_getOpenFile -filetypes {{{MIF files} {.mif}}} \
	    -initialdir $mii_dir -title "Input vector MIF file"]
    if {[string length $s] > 0} then {
	.calc.in.mii_ent delete 0 end
	.calc.in.mii_ent insert 0 $s
	set mii_dir [file dirname $s]
    }
}
label .calc.in.mio_lbl -text "Output MIF file:"
entry .calc.in.mio_ent -relief sunken -bg white -width 40
button .calc.in.mio_but -text ". . ." -font "Helvetica 6" -command {
    set s [tk_getSaveFile -defaultextension .mif \
	    -filetypes {{{MIF files} {.mif}}} \
	    -initialdir $mio_dir -title "Output vector MIF file"]
    if {[string length $s] > 0} then {
	.calc.in.mio_ent delete 0 end
	.calc.in.mio_ent insert 0 $s
	set mio_dir [file dirname $s]
    }
}
grid .calc.in.tab_lbl -row 0 -column 0 -sticky w
grid .calc.in.tab_ent -row 0 -column 1
grid .calc.in.tab_but -row 0 -column 2 -padx 2 -pady 2
grid .calc.in.mii_lbl -row 1 -column 0 -sticky w
grid .calc.in.mii_ent -row 1 -column 1
grid .calc.in.mii_but -row 1 -column 2 -padx 2 -pady 2
grid .calc.in.mio_lbl -row 2 -column 0 -sticky w
grid .calc.in.mio_ent -row 2 -column 1
grid .calc.in.mio_but -row 2 -column 2 -padx 2 -pady 2

frame .calc.coord
radiobutton .calc.coord.mi -text "MapInfo coordinate" -variable coord -value 0
radiobutton .calc.coord.et -text "EasyTrace coordinate" -variable coord -value 1
pack .calc.coord.mi .calc.coord.et -side left

frame .calc.but
button .calc.but.run -text Proceed -width 12 -command {
    if {[string length [.calc.in.tab_ent get]] == 0} then {
	tk_messageBox -icon error -type ok -title "Error" \
		-message "Raster registartion TAB file name must be specified"
    } elseif {[string length [.calc.in.mii_ent get]] == 0} then {
	tk_messageBox -icon error -type ok -title "Error" \
		-message "Input vector MIF file name must be specified"
    } elseif {[string length [.calc.in.mio_ent get]] == 0} then {
	tk_messageBox -icon error -type ok -title "Error" \
		-message "Output vector MIF file name must be specified"
    } elseif {[string compare [.calc.in.mii_ent get] [.calc.in.mio_ent get]] == 0} then {
	tk_messageBox -icon error -type ok -title "Error" \
		-message "Input and output MIF file names must differ"
    } else {
	tk_messageBox -icon info -type ok -title "Success" \
		-message "[eval mifreg $coord]"
    }
}
button .calc.but.about -text About... -width 12 -command {
    if {[winfo exists .about]==0} {
	catch {destroy .about}
	toplevel .about
	wm title .about "About"
	wm resizable .about 0 0
	image create photo splash -file "splash.gif"
	label .about.image -image splash -relief sunken
	label .about.msg1 -justify center \
                -text "The MIF Coordinate Registrator\nVersion 1.10 - 2008-03-20"
	label .about.msg2 -justify center \
                -text "Copyright (C) 2001-2008 by Igor Belov"
	button .about.button -text OK -width 12 -command {
	    wm withdraw .about
	    grab release .about
	}
	pack .about.msg1 .about.image .about.msg2 .about.button \
		-side top -padx .5m -pady .5m
	wm withdraw .about
	update idletasks
	set x [expr {[winfo screenwidth .about]/2 - [winfo reqwidth .about]/2 \
		- [winfo vrootx [winfo parent .about]]}]
	set y [expr {[winfo screenheight .about]/2 - [winfo reqheight .about]/2 \
		- [winfo vrooty [winfo parent .about]]}]
	wm geom .about +$x+$y
    }
    wm deiconify .about
    raise .about
    focus .about
    grab .about
}
button .calc.but.quit -text Quit -width 12 -command {
    destroy .calc
}
pack .calc.but.run .calc.but.about .calc.but.quit \
	-side left -padx 20 -pady 4

pack .calc.msg .calc.in .calc.coord .calc.but

update
focus -force .calc
tkwait window .calc

set fp [open $rc w]
foreach w {tab_dir mii_dir mio_dir coord} {
    puts $fp [format "set $w \"%s\"" [set $w]]
}
close $fp

exit
