#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "plain_c.h"

/* char *tab_hdr[] = {"!table","!version","!charset","","Definition","Type","Fields"}; */
char *tab_hdr[] = {"!table","!version","!charset","definition","file","type","","coordsys","units"};
char *mif_hdr[] = {"Version","Charset","Delimiter","Coordsys","Columns","Data"};
char *token[TOKNUM] = {"None","Point","Line","Pline","Arc","Region","Ellipse","Rect","Roundrect","Text"};
char *mif_atr[] = {"Pen","Brush"};

char *tab_msg[] = {
  "unexpected EOF",
  "unrecognized keyword",
  "points 1,2,3 lay along one line"
};
char *mif_msg[] = {
  "unexpected EOF",
  "unrecognized keyword",
  "CoordSys must be NonEarth",
  "unknown object type",
  "object attribute mismatch",
  "negative number of objects or points"
};

/*-----------------26.09.01 18:28-------------------
 * Convert X[2] coordinate to Y[2] coordinate
 * using affine transform parameters P[2][4]
 * --------------------------------------------------*/
void aff_trans(double *x,double *y,double *p){
  double u[2];
  int j;

  for(j=0;j<2;j++){
    u[j]=x[j]-p[4*j+0];
  }
  for(j=0;j<2;j++){
    y[j]=p[4*j+1]+p[4*j+2]*u[0]+p[4*j+3]*u[1];
  }

  return;
}

/*-----------------22.09.01 21:10-------------------
 * Read raster registration TAB file and compute
 * affine transform parameters, write LOG file.
 * CoordSys is copied to coord_sys.
 * --------------------------------------------------*/
int read_tab_calc_tr_parm(FILE *fp,FILE *log_fp,char *coord_sys,double *p){
  double x[MAXRPR][2],y[MAXRPR][2];
  double dx[2][2],dy[2][2];
  double det;
  double a[2][2],u[2],l[2],q[3],qxl[2][2],qll[2],phi[2],m[2],v[2];
  double ksi[2],rho[2],mu[2],alp[2],the[2];
  char s[MAXSTR],w[2][MAXSTR],*sptr;
  char lbl[MAXRPR][32];
  int stat,i,j,k,n,ipass;

  /* read MapInfo raster definition TAB file */
  stat=n=0;
  p[0]=p[4]=p[1]=p[5]=0.;
  while(stat >= 0){
    if(fgets(s,MAXSTR,fp)==NULL){
      return 1;  /* unexpected EOF */
    }
    w[0][0]=w[1][0]='\0';
    sscanf(s,"%s %s",w[0],w[1]);
    if(strlen(w[0])==0){
      continue;
    }
    for(j=0;j<strlen(w[0]);j++){
      w[0][j]=tolower(w[0][j]);
    }
    if(stat==6){ /* coordinate data records between "Type" and "CoordSys" */

      sptr=strchr(s,'(');
      y[n][0]=strtod(sptr+1,&sptr);
      y[n][1]=strtod(sptr+1,&sptr);
      sptr=strrchr(s,'(');
      x[n][0]=strtod(sptr+1,&sptr);
      x[n][1]=strtod(sptr+1,&sptr);

      p[0]+=(x[n][0]-p[0])/(n+1);
      p[4]+=(x[n][1]-p[4])/(n+1);
      p[1]+=(y[n][0]-p[1])/(n+1);
      p[5]+=(y[n][1]-p[5])/(n+1);

      strcpy(lbl[n],strchr(s,'\"')+1);
      sptr=strchr(lbl[n],'\"');
      sptr[0]='\0';

      sptr=strrchr(s,'\"');             /* last data record */
      if(sptr[1] != ','){
        stat++;
      }

      n++;

    } else {
      if(strcmp(w[0],tab_hdr[stat]) != 0){
        return 2;  /* unrecognized keyword */
      }
      if(stat==8){                      /* "Units" */
        stat=-1;
      } else {
        if(stat==7){                    /* "CoordSys" */
          strcpy(coord_sys,strchr(s,'C'));
        }
        stat++;
      }
    }
  }

  for(i=0;i<2;i++){
    for(j=0;j<2;j++){
      dx[i][j]=x[i+1][j]-x[0][j];
      dy[i][j]=y[i+1][j]-y[0][j];
    }
  }
  det=dx[0][0]*dx[1][1]-dx[1][0]*dx[0][1];
  if(det==0.){
    return 3;                           /* points 1,2,3 lay along one line */
  }
  p[2] = (dx[1][1]*dy[0][0]-dx[0][1]*dy[1][0])/det;
  p[3] =-(dx[1][0]*dy[0][0]-dx[0][0]*dy[1][0])/det;
  p[6] = (dx[1][1]*dy[0][1]-dx[0][1]*dy[1][1])/det;
  p[7] =-(dx[1][0]*dy[0][1]-dx[0][0]*dy[1][1])/det;

  if(n>3){
    for(ipass=0;ipass<NPASS;ipass++){
      q[0]=q[1]=q[2]=0.;
      qxl[0][0]=qxl[0][1]=qxl[1][0]=qxl[1][1]=0.;
      qll[0]=qll[1]=0.;
      for(i=0;i<n;i++){
        for(j=0;j<2;j++){
          u[j]=x[i][j]-p[4*j+0];
        }
        q[0]+=u[0]*u[0];
        q[1]+=u[0]*u[1];
        q[2]+=u[1]*u[1];
        for(j=0;j<2;j++){
          l[j]=p[4*j+1]+p[4*j+2]*u[0]+p[4*j+3]*u[1]-y[i][j];
          for(k=0;k<2;k++){
            qxl[k][j]+=u[k]*l[j];
          }
          qll[j]+=l[j]*l[j];
        }
      }

      det=q[0]*q[2]-q[1]*q[1];
      for(j=0;j<2;j++){
        a[j][0]=(q[1]*qxl[1][j]-q[2]*qxl[0][j])/det;
        a[j][1]=(q[1]*qxl[0][j]-q[0]*qxl[1][j])/det;

        phi[j]=qll[j]+a[j][0]*qxl[0][j]+a[j][1]*qxl[1][j];
        m[j]=sqrt(phi[j]/(n-3));

        p[4*j+2]+=a[j][0];
        p[4*j+3]+=a[j][1];
      }
    }
  }

  fputs("Direct transform\n",log_fp);
  fprintf(log_fp,"A %.16g\n",p[2]);
  fprintf(log_fp,"D %.16g\n",p[6]);
  fprintf(log_fp,"B %.16g\n",p[3]);
  fprintf(log_fp,"E %.16g\n",p[7]);
  fprintf(log_fp,"C %.16g\n",p[1]-(p[2]*p[0]+p[3]*p[4]));
  fprintf(log_fp,"F %.16g\n",p[5]-(p[6]*p[0]+p[7]*p[4]));

  det=p[2]*p[7]-p[3]*p[6];

  fputs("\nInverse transform\n",log_fp);
  fprintf(log_fp,"A' %.16g\n",p[7]/det);
  fprintf(log_fp,"D' %.16g\n",-p[6]/det);
  fprintf(log_fp,"B' %.16g\n",-p[3]/det);
  fprintf(log_fp,"E' %.16g\n",p[2]/det);
  fprintf(log_fp,"C' %.16g\n",p[0]-(p[7]*p[1]-p[3]*p[5])/det);
  fprintf(log_fp,"F' %.16g\n",p[4]-(-p[6]*p[1]+p[2]*p[5])/det);

  ksi[0]=p[2]-p[7];
  ksi[1]=p[3]+p[6];
  rho[0]=sqrt(ksi[0]*ksi[0]+ksi[1]*ksi[1]);
  alp[0]=atan2(ksi[1],ksi[0]);
  ksi[0]=p[2]+p[7];
  ksi[1]=p[3]-p[6];
  rho[1]=sqrt(ksi[0]*ksi[0]+ksi[1]*ksi[1]);
  alp[1]=atan2(ksi[1],ksi[0]);
  mu[0]=(rho[0]+rho[1])/2.;
  mu[1]=(rho[0]-rho[1])/2.;
  the[0]=(alp[0]+alp[1])/2.;
  the[1]=(alp[0]-alp[1])/2.;

  fputs("\nAlternative set\n",log_fp);
  fprintf(log_fp,"Xc %.16g\n",p[0]);
  fprintf(log_fp,"Yc %.16g\n",p[4]);
  fprintf(log_fp,"alpha %+.16g\n",-the[0]*RHO);
  fprintf(log_fp,"mu %.16g\n",mu[0]);
  fprintf(log_fp,"nu %.16g\n",mu[1]);
  fprintf(log_fp,"beta %+.16g\n",-the[1]*RHO);
  fprintf(log_fp,"Xc' %.16g\n",p[1]);
  fprintf(log_fp,"Yc' %.16g\n",p[5]);

  fputs("\nDeviations:\n",log_fp);
  fputs("No\tdX\tdY\tLabel\n\n",log_fp);

  phi[0]=phi[1]=0.;
  for(i=0;i<n;i++){
    for(j=0;j<2;j++){
      u[j]=x[i][j]-p[4*j+0];
    }
    for(j=0;j<2;j++){
      v[j]=y[i][j]-(p[4*j+1]+p[4*j+2]*u[0]+p[4*j+3]*u[1]);
      phi[j]+=v[j]*v[j];
    }
    fprintf(log_fp,"%d.\t%+.3g\t%+.3g\t%s\n",i+1,v[0],v[1],lbl[i]);
  }

  if(n>3){
    for(j=0;j<2;j++){
      m[j]=sqrt(phi[j]/(n-3));
    }
    fputs("\nStd errors:\n",log_fp);
    fprintf(log_fp,"\t%.3g\t%.3g\n",m[0],m[1]);
  }

  return 0;
}

/*-----------------22.09.01 21:19-------------------
 * Read input MIF, transform coordinate, write output MIF file.
 * New CoordSys is taken from coord_sys.
 * --------------------------------------------------*/
int mif_trans_mif(FILE *in_fp,FILE *out_fp,FILE *in_mid,FILE *out_mid,char *coord_sys,double *p,int et){
  double x[2],y[2],bounds[2][2];
  char s[MAXSTR],w[2][MAXSTR],*sptr;
  int stat,i,j,k,n,ntok;

  /* read header of the MapInfo MIF file */
  stat=i=0;
  while(stat >= 0){
    if(fgets(s,MAXSTR,in_fp)==NULL){
      return 1;  /* error reading MapInfo MIF */
    }
    w[0][0]=w[1][0]='\0';
    sscanf(s,"%s %s",w[0],w[1]);
    if(strlen(w[0])==0){
      continue;
    }
    w[0][0]=toupper(w[0][0]);
    for(j=1;j<strlen(w[0]);j++){
      w[0][j]=tolower(w[0][j]);
    }
    if(strcmp(w[0],"Index")==0){
      fputs(s,out_fp);
      continue;
    }
    if(strcmp(w[0],mif_hdr[stat]) != 0){
      return 2;
    }

    switch(stat){
      case 3 :                          /* "CoordSys" */
        w[1][0]=toupper(w[1][0]);
        for(j=1;j<strlen(w[1]);j++){
          w[1][j]=tolower(w[1][j]);
        }
        if(strcmp(w[1],"Nonearth") != 0){
          return 3;
        }

        if((sptr=strchr(s,'h'))==NULL){
          sptr=strchr(s,'H');
        }
        w[0][0]=w[1][0]='\0';
        sscanf(sptr+1,"%s %s",w[0],w[1]);
        if(strlen(w[0])==0){
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          w[0][0]=w[1][0]='\0';
          sscanf(s,"%s %s",w[0],w[1]);
        }
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],"Units") != 0){
          return 2;
        }

        sptr=strrchr(s,'\"');
        w[0][0]=w[1][0]='\0';
        sscanf(sptr+1,"%s %s",w[0],w[1]);
        if(strlen(w[0])==0){
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          w[0][0]=w[1][0]='\0';
          sscanf(s,"%s %s",w[0],w[1]);
        }
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],"Bounds") != 0){
          return 2;
        }

        sptr=strchr(s,'(');
        bounds[0][0] = strtod(sptr+1,&sptr);
        bounds[0][1] = strtod(sptr+1,&sptr);
        sptr=strrchr(s,'(');
        bounds[1][0] = strtod(sptr+1,&sptr);
        bounds[1][1] = strtod(sptr+1,&sptr);

        fputs(coord_sys,out_fp);
        stat++;
        break;
      case 4 :                          /* "Columns" */
        n=atoi(w[1]);
        fprintf(out_fp,"%s %d\n",w[0],n);
        for(i=0;i<n;i++){
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          fputs(s,out_fp);
        }
        stat++;
        break;
      case 5 :                          /* "Data" */
        fprintf(out_fp,"%s\n\n",w[0]);
        stat=-1;
        break;
      default:
        fprintf(out_fp,"%s %s\n",w[0],w[1]);
        stat++;
        break;
      /* end of switch */
    }
  }

  /* handle data */
  while(fgets(s,MAXSTR,in_fp)!=NULL){
    w[0][0]=w[1][0]='\0';
    sscanf(s,"%s %s",w[0],w[1]);
    if(strlen(w[0])==0){
      continue;
    }
    w[0][0]=toupper(w[0][0]);
    for(j=1;j<strlen(w[0]);j++){
      w[0][j]=tolower(w[0][j]);
    }
    if(strcmp(w[0],"Smooth")==0){
      fputs(s,out_fp);
      continue;
    }

    /* get token */
    for(ntok = 0; ntok < TOKNUM; ntok++) {
      if(strcmp(w[0],token[ntok]) == 0) {
        break;
      }
    }
    if(ntok == TOKNUM) {
      return 4;
    }

    switch(ntok) {
      case 0 :                    /* none */
	fputs("none\n",out_fp);
        if(in_mid!=NULL && fgets(s,MAXSTR,in_mid)!=NULL && out_mid!=NULL){
          fputs(s,out_mid);
        }
        break;
      case 1 :                    /* Point */
        x[0]=strtod(s+strlen(w[0])+1,&sptr);
	x[1]=strtod(sptr,&sptr);
	if(et!=0)
	  x[1]=bounds[1][1]-x[1];
        aff_trans(x,y,p);
        fprintf(out_fp,"%s %.9g %.9g\n",token[ntok],y[0],y[1]);
        if(fgets(s,MAXSTR,in_fp)==NULL){
          return 1;
        }
        w[0][0]=w[1][0]='\0';
        sscanf(s,"%s %s",w[0],w[1]);
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],"Symbol")!=0){
          return 5;
        }
	fputs(s,out_fp);

        if(in_mid!=NULL && fgets(s,MAXSTR,in_mid)!=NULL && out_mid!=NULL){
          fputs(s,out_mid);
        }
        break;
      case 2 :                    /* Line */
        x[0]=strtod(s+strlen(w[0])+1,&sptr);
	x[1]=strtod(sptr,&sptr);
	if(et!=0)
	  x[1]=bounds[1][1]-x[1];
        aff_trans(x,y,p);
        fprintf(out_fp,"%s %.9g %.9g ",token[ntok],y[0],y[1]);
        x[0]=strtod(sptr,&sptr);
	x[1]=strtod(sptr,&sptr);
	if(et!=0)
	  x[1]=bounds[1][1]-x[1];
        aff_trans(x,y,p);
        fprintf(out_fp,"%.9g %.9g\n",y[0],y[1]);
        if(fgets(s,MAXSTR,in_fp)==NULL){
          return 1;
        }
        w[0][0]=w[1][0]='\0';
        sscanf(s,"%s %s",w[0],w[1]);
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],mif_atr[0])!=0){
          return 5;
        }
	fputs(s,out_fp);

        if(in_mid!=NULL && fgets(s,MAXSTR,in_mid)!=NULL && out_mid!=NULL){
          fputs(s,out_mid);
        }
        break;
      case 3 :                    /* Pline */
        w[1][0]=toupper(w[1][0]);
        for(j=1;j<strlen(w[1]);j++){
          w[1][j]=tolower(w[1][j]);
        }
        if(strcmp(w[1],"Multiple") == 0){ /* multiple Pline */
          k=atoi(s+strlen(w[0])+strlen(w[1])+2);
          if(k < 1) {
            return 6;
          }
          fprintf(out_fp,"%s %s %d\n",w[0],w[1],k);
          for(j=0;j<k;j++) {
            if(fgets(s,MAXSTR,in_fp) == NULL) {
              return 1;
            }
            n=atoi(s);
            if(n < 1){
              return 6;
            }
            fprintf(out_fp,"  %d\n",n);
            for(i=0;i<n;i++){
              if(fgets(s,MAXSTR,in_fp) == NULL) {
                return 1;
              }
              sscanf(s,"%lf %lf",&x[0],&x[1]);
	      if(et!=0)
		x[1]=bounds[1][1]-x[1];
              aff_trans(x,y,p);
              fprintf(out_fp,"%.9g %.9g\n",y[0],y[1]);
            }
          }
        } else {                        /* single Pline */
          n = atoi(w[1]);
          if(n < 1) {
            return 6;
          }
          fprintf(out_fp,"%s %d\n",w[0],n);
          for(i=0;i<n;i++){
            if(fgets(s,MAXSTR,in_fp) == NULL) {
              return 1;
            }
            sscanf(s,"%lf %lf",&x[0],&x[1]);
	    if(et!=0)
	      x[1]=bounds[1][1]-x[1];
            aff_trans(x,y,p);
            fprintf(out_fp,"%.9g %.9g\n",y[0],y[1]);
          }
        }
        if(fgets(s,MAXSTR,in_fp)==NULL){
          return 1;
        }
        w[0][0]=w[1][0]='\0';
        sscanf(s,"%s %s",w[0],w[1]);
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],mif_atr[0])!=0){
          return 5;
        }
	fputs(s,out_fp);

        if(in_mid!=NULL && fgets(s,MAXSTR,in_mid)!=NULL && out_mid!=NULL){
          fputs(s,out_mid);
        }
        break;
      case 4 :                          /* Arc */
        if(fgets(s,MAXSTR,in_fp)==NULL){
          return 1;
        }
        w[0][0]=w[1][0]='\0';
        sscanf(s,"%s %s",w[0],w[1]);
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],mif_atr[0])!=0){
          return 5;
        }

        if(in_mid!=NULL){
	  fgets(s,MAXSTR,in_mid);
	}

        break;
      case 5 :                          /* Region */
        k=atoi(w[1]);
        if(k<1){
          return 6;
        }
        fprintf(out_fp,"%s %d\n",w[0],k);
        for(j=0;j<k;j++){
          if(fgets(s,MAXSTR,in_fp) == NULL){
            return 1;
          }
          n=atoi(s);
          if(n<1){
            return 6;
          }
          fprintf(out_fp,"  %d\n",n);
          for(i=0;i<n;i++){
            if(fgets(s,MAXSTR,in_fp) == NULL) {
              return 1;
            }
            sscanf(s,"%lf %lf",&x[0],&x[1]);
	    if(et!=0)
	      x[1]=bounds[1][1]-x[1];
            aff_trans(x,y,p);
            fprintf(out_fp,"%.9g %.9g\n",y[0],y[1]);
          }
        }
        for(i=0;i<2;i++){
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          w[0][0]=w[1][0]='\0';
          sscanf(s,"%s %s",w[0],w[1]);
          w[0][0]=toupper(w[0][0]);
          for(j=1;j<strlen(w[0]);j++){
            w[0][j]=tolower(w[0][j]);
          }
          if(strcmp(w[0],mif_atr[i])!=0){
            return 5;
          }
	  fputs(s,out_fp);
        }
        if(fgets(s,MAXSTR,in_fp)==NULL){
          return 1;
        }
        w[0][0]='\0';
        sscanf(s,"%s %lf %lf",w[0],&x[0],&x[1]);
        w[0][0]=toupper(w[0][0]);
        for(j=1;j<strlen(w[0]);j++){
          w[0][j]=tolower(w[0][j]);
        }
        if(strcmp(w[0],"Center")!=0){
          return 5;
        }
	if(et!=0)
	  x[1]=bounds[1][1]-x[1];
        aff_trans(x,y,p);
        fprintf(out_fp,"    %s %.9g %.9g\n",w[0],y[0],y[1]);

        if(in_mid!=NULL && fgets(s,MAXSTR,in_mid)!=NULL && out_mid!=NULL){
          fputs(s,out_mid);
        }
        break;
      case 9 :                          /* Text */
        do{
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          w[0][0]='\0';
          sscanf(s,"%s",w[0]);
        } while(strlen(w[0]) != 0);

	if(in_mid!=NULL){
	  fgets(s,MAXSTR,in_mid);
	}

        break;
      default:                          /* "Ellipse","Rect","Roundrect" */
        for(i=0;i<2;i++){
          if(fgets(s,MAXSTR,in_fp)==NULL){
            return 1;
          }
          w[0][0]=w[1][0]='\0';
          sscanf(s,"%s %s",w[0],w[1]);
          w[0][0]=toupper(w[0][0]);
          for(j=1;j<strlen(w[0]);j++){
            w[0][j]=tolower(w[0][j]);
          }
          if(strcmp(w[0],mif_atr[i])!=0){
            return 5;
          }
        }

	if(in_mid!=NULL){
	  fgets(s,MAXSTR,in_mid);
	}

        break;
      /* end of switch */
    }
  }

  return 0;
}
